
# -- import packages: ---------------------------------------------------------
import torch


# -- import local dependencies: -----------------------------------------------
from .core._base_neural_ode import BaseODE


# -- import standard libraries and define types: ------------------------------
from typing import Union, List


# -- Main operational class: --------------------------------------------------
class NeuralODE(BaseODE):
    DIFFEQ_TYPE = "ODE"
    
    """NeuralODE"""
    
    def __init__(
        self,
        state_size: int,
        dt: float = 0.1,
        coef_diff: float = 0,
        mu_hidden: Union[List[int], int] = [512, 512],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        sde_type: str = "ito",
        noise_type: str = "general",
    ) -> None:
        """
        Args:
            state_size: int
            
            dt: float = 0.1
            
            coef_diff: float = 0
            
            mu_hidden: Union[List[int], int] = [512, 512]
            
            mu_activation: Union[str, List[str]] = "LeakyReLU"
            
            mu_dropout: Union[float, List[float]] = 0.
            
            mu_bias: bool = True
            
            mu_output_bias: bool = True
            
            mu_n_augment: int = 0
            
            sde_type: str = "ito"
            
            noise_type (str): Description. **Default**: ``"general"``.
            
        Returns:
            None
        """

        # -- auto: we are not using the diffusion: ----------------------------

        sigma_hidden = []
        sigma_output_bias = False
        brownian_dim = 1

        super().__init__()

        self.__config__(locals())

    def drift(self, y: torch.Tensor) -> torch.Tensor:
        """ """
        return self.mu(y)

    def forward(self, t, y: torch.Tensor) -> torch.Tensor:
        """ """
        return self.mu(y)




# -- import packages: ---------------------------------------------------------
import torch


# -- import local dependencies: ----------------------------------------------
from . import core


# -- import standard libraries and define types: -----------------------------
from typing import Union, List


# -- Main operational class: -------------------------------------------------
class LatentPotentialODE(core.BaseLatentODE):
    DIFFEQ_TYPE = "ODE"

    def __init__(
        self,
        state_size: int,
        coef_diff: float = 0,
        dt: float = 0.1,
        mu_hidden: Union[List[int], int] = [512, 512],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.,
        mu_bias: Union[bool, List[bool]] = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        sde_type: str = "ito",
        noise_type: str = "general",
        brownian_dim: int = 1,
    ):
        """Instantiate a LatentPotentialODE
        
        Args:
            state_size (int): Description.
            
            coef_diff (float): Description. ``**Default**: 0``
            
            dt (float): Description. ``**Default**: 0.1``
            
            mu_hidden: Union[List[int], int] = [512, 512].
            
            mu_activation: Union[str, List[str]] = "LeakyReLU".
            
            mu_dropout: Union[float, List[float]] = 0..
            
            mu_bias: Union[bool, List[bool]] = True.
            
            mu_output_bias: bool = True.
            
            mu_n_augment: int = 0.
            
            sde_type: str = "ito".
            
            noise_type: str = "general".
            
            brownian_dim: int = 1.
            
        Returns:
            None
            
        Notes:
            By declaring ``mu_potential = False``, we explicitly state that
            we are not are not learning a potential function by default.
        """
        
        super().__init__()

        mu_potential = False
        self.__config__(locals())
        self.potential = core.Potential(state_size = self._state_size)

    def drift(self, y: torch.Tensor) -> torch.Tensor:
        """
        Args:
            y (torch.Tensor): input state.
            
        Returns:
            y_hat_f (torch.Tensor): drift(input).
        """
        return self.mu(y)

    def prior_drift(self, y: torch.Tensor) -> torch.Tensor:
        """
        Args:
            y (torch.Tensor): input state.
            
        Returns:
            y_hat_h (torch.Tensor): prior_drift(input).
        """
        return self.potential.potential_gradient(y)




# -- import packages: ----------------------------------------------------------
import torch


# -- import local dependencies: ------------------------------------------------
from . import core


# -- import standard libraries and define types: -------------------------------
from typing import Union, List


# -- Main operational class: ---------------------------------------------------
class PotentialSDE(core.BaseSDE):
    """PotentialSDE, subclass of core.BaseSDE"""
    DIFFEQ_TYPE = "SDE"
    def __init__(
        self,
        state_size: int,
        mu_hidden: Union[List[int], int] = [512, 512],
        sigma_hidden: Union[List[int], int] = [32, 32],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        sigma_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.,
        sigma_dropout: Union[float, List[float]] = 0.,
        mu_bias: Union[bool, List[bool]] = True,
        sigma_bias: Union[bool, List[bool]] = True,
        sigma_output_bias: bool = True,
        mu_n_augment: int = 0,
        sigma_n_augment: int = 0,
        sde_type: str = "ito",
        noise_type: str = "general",
        brownian_dim: int = 1,
        coef_drift: float = 1.,
        coef_diffusion: float = 1.,
    ):
        
        """PotentialSDE instantiation from parameters are parsed and passed to the base class config function.
        
        Args:
            state_size (int): Input and output state size of the differential equation.
        
            mu_hidden (Union[List[int], int]): Architecture of the hidden layers of the drift neural network. **Default**: ``[512, 512]``.
        
            sigma_hidden (Union[List[int], int]): Architecture of the hidden layers of the diffusion neural network. **Default**: ``[32, 32]``.
            
            mu_activation (Union[str, List[str]]): Activation function(s) used in each layer of the drift neural network. If ``len(mu_hidden) > len(mu_activation)``, the remaining activation functions are autofilled using the last value passed. **Default**: ``"LeakyReLU"``.
            
            sigma_activation (Union[str, List[str]]): Activation function(s) used in each layer of the diffusion neural network. If ``len(sigma_hidden) > len(sigma_activation)``, the remaining activation functions are autofilled using the last value passed. **Default**: ``"LeakyReLU"``.
            
            mu_dropout (Union[float, List[float]]): Description. **Default**: ``0.``.
            
            sigma_dropout (Union[float, List[float]]): Description. **Default**: ``0.``.
            
            mu_bias (Union[bool, List[bool]]): Description. **Default**: ``True``.
            
            sigma_bias (Union[bool, List[bool]]): Description. **Default**: ``True``.
            
            mu_output_bias (bool): Description. **Default**: ``False``.
            
            sigma_output_bias (bool): Description. **Default**: ``True``.
            
            mu_n_augment (int): Description. **Default**: ``0``.
            
            sigma_n_augment (int): Description. **Default**: ``0``.
            
            sde_type (str): Description. **Default**: ``"ito"``.
            
            noise_type (str): Description. **Default**: ``"general"``.
            
            brownian_dim (int): Number of diffusion dimensions. **Default**: ``1``.
            
            coef_drift (float): Multiplier of drift network output. **Default**: ``1.``.
            
            coef_diffusion (float): Multiplier of diffusion network output. **Default**: ``1.``.
        
        Returns:
            None
        
        Notes:
            NeuralSDE: torch.nn.Module
            
        Example:
        
            .. code-block:: python

                import neural_diffeqs
                SDE = neural_diffeqs.PotentialSDE(state_size = 50)
        """
        
        super().__init__()

        mu_potential = True
        self.__config__(locals())
        
    def _potential(self, y: torch.Tensor) -> torch.Tensor:
        
        """ """
        
        y = y.requires_grad_()
        return self.mu(y)

    def _gradient(self, ψ: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        
        """use-case: output is directly psi (from a potential network)"""
        
        return torch.autograd.grad(ψ, y, torch.ones_like(ψ), create_graph=True)[0]

    def drift(self, y: torch.Tensor) -> torch.Tensor:
        
        """ """
        
        y = y.requires_grad_()
        ψ = self._potential(y)
        return self._gradient(ψ, y) * self._coef_drift

    def diffusion(self, y: torch.Tensor) -> torch.Tensor:
        
        """ """
        
        return self.sigma(y).view(y.shape[0], y.shape[1], self._brownian_dim) * self._coef_diffusion




# -- import packages: ---------------------------------------------------------
import torch
import ABCParse


# -- import local dependencies: -----------------------------------------------
from .core._base_neural_sde import BaseSDE


# -- import standard libraries and define types: ------------------------------
from typing import Union, List, Any


# -- Main operational class: --------------------------------------------------
class NeuralSDE(BaseSDE):
    DIFFEQ_TYPE = "SDE"
    """NeuralSDE - a neural differential equation."""
    def __init__(
        self,
        state_size: int,
        mu_hidden: Union[List[int], int] = [512, 512],
        sigma_hidden: Union[List[int], int] = [32, 32],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        sigma_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.,
        sigma_dropout: Union[float, List[float]] = 0.,
        mu_bias: Union[bool, List[bool]] = True,
        sigma_bias: Union[bool, List[bool]] = True,
        mu_output_bias: bool = True,
        sigma_output_bias: bool = True,
        mu_n_augment: int = 0,
        sigma_n_augment: int = 0,
        sde_type: str = "ito",
        noise_type: str = "general",
        brownian_dim: int = 1,
        coef_drift: float = 1.,
        coef_diffusion: float = 1.,
    ) -> None:
        
        """NeuralSDE instantiation from parameters are parsed and passed to the base class config function.
        
        Args:
            state_size (int): Input and output state size of the differential equation.
        
            mu_hidden (Union[List[int], int]): Architecture of the hidden layers of the drift neural network. **Default**: ``[512, 512]``.
        
            sigma_hidden (Union[List[int], int]): Architecture of the hidden layers of the diffusion neural network. **Default**: ``[32, 32]``.
            
            mu_activation (Union[str, List[str]]): Activation function(s) used in each layer of the drift neural network. If ``len(mu_hidden) > len(mu_activation)``, the remaining activation functions are autofilled using the last value passed. **Default**: ``"LeakyReLU"``.
            
            sigma_activation (Union[str, List[str]]): Activation function(s) used in each layer of the diffusion neural network. If ``len(sigma_hidden) > len(sigma_activation)``, the remaining activation functions are autofilled using the last value passed. **Default**: ``"LeakyReLU"``.
            
            mu_dropout (Union[float, List[float]]): Description. **Default**: ``0.``.
            
            sigma_dropout (Union[float, List[float]]): Description. **Default**: ``0.``.
            
            mu_bias (Union[bool, List[bool]]): Description. **Default**: ``True``.
            
            sigma_bias (Union[bool, List[bool]]): Description. **Default**: ``True``.
            
            mu_output_bias (bool): Description. **Default**: ``True``.
            
            sigma_output_bias (bool): Description. **Default**: ``True``.
            
            mu_n_augment (int): Description. **Default**: ``0``.
            
            sigma_n_augment (int): Description. **Default**: ``0``.
            
            sde_type (str): Description. **Default**: ``"ito"``.
            
            noise_type (str): Description. **Default**: ``"general"``.
            
            brownian_dim (int): Number of diffusion dimensions. **Default**: ``1``.
            
            coef_drift (float): Multiplier of drift network output. **Default**: ``1.``.
            
            coef_diffusion (float): Multiplier of diffusion network output. **Default**: ``1.``.
        
        Returns:
            None
        
        Notes:
            NeuralSDE: torch.nn.Module
            
        Example:
        
            .. code-block:: python

                import neural_diffeqs
                SDE = neural_diffeqs.NeuralSDE(state_size = 50)
        """
        super().__init__()

        self.__config__(locals())

    def drift(self, y: torch.Tensor) -> torch.Tensor:
        """Drift function.
        
        Args:
            y (torch.Tensor): input state.
            
        Returns:
            torch.Tensor: drift(y)
            
        Example:
        
            .. code-block:: python
            
                y_hat_f = SDE.drift(y = y)
        """
        return self.mu(y) * self._coef_drift

    def diffusion(self, y: torch.Tensor) -> torch.Tensor:
        """Diffusion function.
        
        Args:
            y (torch.Tensor): input state.
            
        Returns:
            torch.Tensor: diffusion(y)
            
            
        Example:
        
            .. code-block:: python
                
                y_hat_g = SDE.diffusion(y = y)
        """
        return self.sigma(y).view(y.shape[0], y.shape[1], self._brownian_dim) * self._coef_diffusion




# -- import packages: ---------------------------------------------------------
import torch


# -- import local dependencies: -----------------------------------------------
from .core import BaseODE


# -- import standard libraries and define types: ------------------------------
from typing import Union, List


# -- Main operational class: --------------------------------------------------
class PotentialODE(BaseODE):
    DIFFEQ_TYPE = "ODE"
    
    """PotentialODE"""
    
    def __init__(
        self,
        state_size: int,
        dt: float = 0.1,
        coef_diff: float = 0,
        mu_hidden: Union[List[int], int] = [512, 512],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        sde_type: str = "ito",
        noise_type: str = "general",
        brownian_dim: int = 1,
    ) -> None:
        
        """
        Args:
            state_size: int
            dt: float = 0.1
            coef_diff: float = 0
            mu_hidden: Union[List[int], int] = [512, 512]
            mu_activation: Union[str, List[str]] = "LeakyReLU"
            mu_dropout: Union[float, List[float]] = 0.
            mu_bias: bool = True
            mu_output_bias: bool = True
            mu_n_augment: int = 0
            sde_type: str = "ito"
            noise_type: str = "general"
            brownian_dim: int = 1
            
        Returns:
            None
        """
        
        super().__init__()
        
        mu_potential = True

        self.__config__(locals())

    def _potential(self, y: torch.Tensor) -> torch.Tensor:
        """ """
        return self.mu(y)

    def _gradient(self, ψ: torch.Tensor, y:  torch.Tensor) -> torch.Tensor:
        """use-case: output is directly psi (from a potential network)"""
        return torch.autograd.grad(ψ, y, torch.ones_like(ψ), create_graph=True)[0]

    def drift(self, y: torch.Tensor) -> torch.Tensor:
        """ """
        y = y.requires_grad_()
        ψ = self._potential(y)
        return self._gradient(ψ, y)
